#include "tx_coo_matrix.h"
#include "tx_csr_matrix.h"
#include "tx_dia_matrix.h"
#include "tx_ell_matrix.h"
#include "tx_hyb_matrix.h"

#ifndef GPU_TIMER
#define GPU_TIMER
#endif 
#undef GPU_TIMER

//////////////////////////////////////////
// Interface class
//////////////////////////////////////////
class GPU_Matrix_Ifc {
public:

  /* Constructor */
  GPU_Matrix_Ifc() {}

  /* Destructor */
  virtual ~GPU_Matrix_Ifc() {
  }

  /* Set the Matrix */
  virtual void setMatrix(const cusp::csr_matrix<PetscInt,PetscScalar,cusp::host_memory>& matCPU) = 0;

  /* Set the Matrix */
  virtual void setMatrix(const PetscInt numRows, const PetscInt numCols, const PetscInt nz,
			 const PetscInt * row_offsets, const PetscInt * column_indices,
			 const PetscScalar * values) = 0;
  
  /* Set the Indices For cprow==true (compressed row), m = number of rows */
  virtual void setCPRowIndices(const PetscInt * indices, const int m) = 0;

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const PetscInt * inodes, const int inodeCountPlusOne, const int nnzPerRowMax,
                            const int nodeMax, const int nodeCount) = 0;

  /* build streams and events */
  virtual int buildStreamsAndEvents(const PetscInt MPICommSize, cudaStream_t * theBodyStream) = 0;

  /* needed test for synchronization */
  virtual bool hasNonZeroStream() = 0;

  /* Multiply */
  virtual int multiply(const PetscScalar * x, PetscScalar *y) = 0;

  /* Multiply Add */
  virtual int multiplyAdd(const PetscScalar * x, PetscScalar *y) = 0;
};

////////////////////////////////////////////////////
// Derived class templated over matrix type
////////////////////////////////////////////////////
template<class MTYPE>
class GPU_Matrix : public GPU_Matrix_Ifc {
public:
  
  /* Constructor */
  GPU_Matrix() {
    matClass = new MTYPE;

    /* stream stuff */
    stream= 0;
    streamIsACopy=false;
#ifdef GPU_TIMER
    startGPU=0;
    stopGPU=0;
    dt = 0.f;
    N = 0;
#endif
  }

  /* Destructor */
  virtual ~GPU_Matrix() {
    if (stream && !streamIsACopy) {
      cudaError_t err = cudaStreamDestroy(stream);
      if (err!=cudaSuccess)
	printf("Failed in ~GPU_Matrix() ... cudaStreamDestroy\n");
      stream=0;
    }
#ifdef GPU_TIMER
    printf("MatMult(Add) Mean dt = %1.5f\t N=%d\n",dt/N, N);    

    // Kill the events
    if (startGPU) {
      cudaError_t err = cudaEventDestroy(startGPU);
      if (err!=cudaSuccess)
	printf("Failed in ~GPU_Matrix() ... cudaEventDestroy ... %s\n",cudaGetErrorString(err));
      startGPU=0;
    }
    if (stopGPU) {
      cudaError_t err = cudaEventDestroy(stopGPU);
      if (err!=cudaSuccess)
	printf("Failed in ~GPU_Matrix() ... cudaEventDestroy ... %s\n",cudaGetErrorString(err));
      stopGPU=0;
    }
#endif
    delete matClass;
  }

  /* Set the Matrix */
  void setMatrix(const cusp::csr_matrix<PetscInt,PetscScalar,cusp::host_memory>& matCPU) {
    matClass->setMatrix(matCPU);
  }

  /* Set the Matrix */
  void setMatrix(const PetscInt numRows, const PetscInt numCols, const PetscInt nz,
		 const PetscInt * row_offsets, const PetscInt * column_indices,
		 const PetscScalar * values) {
    
    // figure out in the class that we templated over
    matClass->setMatrix(numRows, numCols, nz, row_offsets, column_indices, values);
  }

  /* Set the Indices For cprow==true (compressed row), m = number of rows */
  void setCPRowIndices(const PetscInt * indices, const int m) {
    matClass->setCPRowIndices(indices, m);
  }

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const PetscInt * inodes, const int inodeCountPlusOne, const int nnzPerRowMax,
                            const int nodeMax, const int nodeCount) {
    matClass->setInodeData(inodes, inodeCountPlusOne, nnzPerRowMax, nodeMax, nodeCount);
  }

  /* build streams and events */
  virtual int buildStreamsAndEvents(const PetscInt MPICommSize, cudaStream_t * theStream) {
    if (MPICommSize>1 && !(*theStream)) {
      cudaError_t err = cudaStreamCreate(&stream);
      *theStream = stream;
      if (err!=cudaSuccess) {
	printf("Failed in GPU_Matrix::buildStreamsAndEvents() ... cudaStreamCreate\n");
	return 1;
      }
    }
    else {
      // set to stream 0 if only 1 MPI rank or the input stream is 0.
      stream = *theStream;
      streamIsACopy=true;
    }
#ifdef GPU_TIMER
    cudaError_t err = cudaEventCreate(&startGPU);
    if (err!=cudaSuccess) {
      printf("Failed in GPU_Matrix::buildStreamsAndEvents() ... cudaEventCreate ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    err = cudaEventCreate(&stopGPU);
    if (err!=cudaSuccess) {
      printf("Failed in GPU_Matrix::buildStreamsAndEvents() ... cudaEventCreate ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    dt=0.f;
    N=0;
#endif
    // success
    return 0;
  };

  /* needed test for synchronization */
  virtual bool hasNonZeroStream() {
    if (stream)
      return true;
    else 
      return false;
  }

  /* Multiply */
  int multiply(const PetscScalar * x, PetscScalar *y) {

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif

    int err = matClass->multiply(stream, x, y);

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float deltaT=0.;
    cudaEventElapsedTime(&deltaT, startGPU, stopGPU);	    
    dt+=deltaT;
    N++;
#endif
    return err;
  };

  /* Multiply Add */
  int multiplyAdd(const PetscScalar * x, PetscScalar *y) {
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif

    int err= matClass->multiplyAdd(stream, x, y);

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float deltaT=0.;
    cudaEventElapsedTime(&deltaT, startGPU, stopGPU);	    
    dt+=deltaT;
    N++;
#endif
   return err;
   
  };

private:
  MTYPE*  matClass;
  typename MTYPE::MatrixType mat;
  cudaStream_t      stream; /*stream for running kernels that overlap with communication */
  bool streamIsACopy;
#ifdef GPU_TIMER
  cudaEvent_t      startGPU;
  cudaEvent_t      stopGPU;
  float            dt;
  PetscInt         N; 
#endif

};

////////////////////////////////////////////////////
// Factory Class
////////////////////////////////////////////////////
class GPU_Matrix_Factory {
public:
  static GPU_Matrix_Ifc *getNew(std::string type){
    if (type == "coo"){
      return new GPU_Matrix<CooMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "csr"){
      return new GPU_Matrix<CsrMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "dia"){
      return new GPU_Matrix<DiaMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "ell"){
      return new GPU_Matrix<EllMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "hyb"){
      return new GPU_Matrix<HybMatrix<PetscInt, PetscScalar> >();
    }
    else {
      // default to CSR
      return new GPU_Matrix<CsrMatrix<PetscInt, PetscScalar> >();
    }
  }
};

